/******************************************************************************
*
* project name:    TIGCC Tools Suite
* file name:       tribute.c
* initial date:    17/08/2000
* author:          thomas.nussbaumer@gmx.net
* description:     A Tribute to Gary Larson (demo of TTArchive functions)
*
* $Id: tribute.c,v 1.4 2002/03/14 10:41:17 tnussb Exp $
*
******************************************************************************/

#define SAVE_SCREEN
#define OPTIMIZE_ROM_CALLS
#define NO_EXIT_SUPPORT
#define USE_TI89
#define USE_TI92P
#define USE_V200

#include <tigcclib.h>

#include "../../src/calctools/ttunpack.h"
#include "../../src/calctools/ttarchive.h"


unsigned char *plane1 = 0;
unsigned char *plane2 = 0;

//=============================================================================
// fetch picture from ttarchive and decompress it to buffer
//=============================================================================
static inline short GetPic(unsigned char* buffer, unsigned short index,unsigned char* archive) {
    unsigned short entries;
    unsigned short i;

    if (!ttarchive_valid(archive)) {
        printf("not an archive\n[press a key]");
        ngetchx();
        return 0;
    }

    entries = ttarchive_entries(archive);

    for (i=0;i<entries;i++) {
        short retval;
        if (i==index) {
            retval = ttunpack_decompress(ttarchive_data(archive,i),buffer);

            if (!retval) return 1;
            else         return 0;
        }
    }

    return 0;
}

//=============================================================================
// show the picture
//=============================================================================
static inline short ShowPic(unsigned char* buffer) {
    short          i;
    unsigned char* dest1 = plane1;
    unsigned char* dest2 = plane2;
    unsigned char* src1  = buffer;
    unsigned char* src2  = buffer+2000;

    if (!TI89) {
        dest1 += 30*14 + 5;
        dest2 += 30*14 + 5;
    }

    for (i=0;i<100;i++) {
        memcpy(dest1,src1,20);
        memcpy(dest2,src2,20);
        dest1+=30;
        dest2+=30;
        src1+=20;
        src2+=20;
    }

    return ngetchx();
}

//=============================================================================
// nomen omen est
//=============================================================================
void _main(void) {
    unsigned char* src;
    SYM_ENTRY*     symptr = DerefSym(SymFind($(gary)));
    unsigned char  buffer[4000];
    short          i;

    MoveTo(0,0);

    if(!symptr) {
        printf("data file not found\n[press a key]");
        ngetchx();
        return;
    }

    src = HLock(symptr->handle);

    src+=2;  // skip "length of variable" bytes

    //---------------------------
    // turn on grayscale graphics
    //---------------------------
    if (!GrayOn()) {
        printf("graymode failure\n[press a key]");
        HeapUnlock(symptr->handle);
        ngetchx();
        return;
    }

    //-----------------------------------------
    // setup bitplane globals and show the logo
    //-----------------------------------------
    plane1 = GetPlane(0);
    plane2 = GetPlane(1);

    memset(plane1,0,30*128);
    memset(plane2,0,30*128);


    for (i=0;i<15;i++) {
        if (GetPic(buffer,i,src)) {
            if (ShowPic(buffer) == 264) break;
        }
        else {
            printf("no pic with index %d\n[press a key]",i);
            ngetchx();
            break;
        }
    }

    GrayOff();
    HeapUnlock(symptr->handle);
}

//#############################################################################
//###################### NO MORE FAKES BEYOND THIS LINE #######################
//#############################################################################
//
//=============================================================================
// Revision History
//=============================================================================
//
// $Log: tribute.c,v $
// Revision 1.4  2002/03/14 10:41:17  tnussb
// source "modernized" for TIGCC v0.93 and above
//
// Revision 1.3  2001/02/10 20:24:16  Thomas Nussbaumer
// using now new calctools header files
//
// Revision 1.2  2000/08/20 16:14:38  Thomas Nussbaumer
// (1) using now shared unpack.c
// (2) using now ttarchive.h
// (3) HeapUnlock() added before returning from main
//
// Revision 1.1  2000/08/16 23:14:08  Thomas Nussbaumer
// initial version
//
//
//
