/******************************************************************************
*
* project name:    TIGCC Tools Suite
* file name:       ttarchive.h
* initial date:    20/08/2000
* author:          thomas.nussbaumer@gmx.net
* description:     structures, definitions and macros to use ttarchive
*
* $Id: ttarchive.h,v 1.4 2002/02/07 09:35:11 tnussb Exp $
*
******************************************************************************/

#ifndef __TTARCHIVE_H__
#define __TTARCHIVE_H__

//-----------------------------------------------------------------------------
// entry of archive structure
//-----------------------------------------------------------------------------
typedef struct {
   unsigned char offset[2];  // offset to the entry data (high byte first)
   unsigned char length[2];  // length of entry
   char          name[8];    // entry name
   unsigned char misc1[2];   // info from cfg file (may be queried by a program)
   unsigned char misc2[2];   // info from cfg file (may be queried by a program)
} TTAEntry;


//-----------------------------------------------------------------------------
// smart macros to access ttarchive
//
// pa    ... pointer to archive start address
// idx   ... index of entry
// entry ... pointer to TTAEntry structure
//
// NOTE: No checking is done in the macros !!
//-----------------------------------------------------------------------------

#define IsTTArchive(pa)         (*(pa) == 't' && *((pa)+1) == 't' && *((pa)+2) == 'a' && *((pa)+3) == 0)
#define GetNrEntries(pa)        ((unsigned int)(*((pa)+4) << 8) | (unsigned int)(*((pa)+5)))
#define GetEntryInfo(pa,idx)    (TTAEntry*)((pa) + 6 + sizeof(TTAEntry)*(unsigned int)(idx))
#define GetEntryStart(pa,entry) ((pa) + 6 + (unsigned int)(sizeof(TTAEntry)*GetNrEntries(pa)) + \
                                (((unsigned int)entry->offset[0]) << 8 | (unsigned int)entry->offset[1]))
#define GetEntrySize(entry)     (((unsigned int)entry->length[0]) << 8 | (unsigned int)entry->length[1])

#define TTA_DESC_LENGTH 20

#endif

//#############################################################################
//###################### NO MORE FAKES BEYOND THIS LINE #######################
//#############################################################################
//
//=============================================================================
// Revision History
//=============================================================================
//
// $Log: ttarchive.h,v $
// Revision 1.4  2002/02/07 09:35:11  tnussb
// Macro GetEntrySize added (used in ttunarchive)
//
// Revision 1.3  2000/11/18 16:05:08  Thomas Nussbaumer
// bug in GetEntryStart macro fixed
//
// Revision 1.2  2000/08/23 01:05:31  Thomas Nussbaumer
// TTA_DESC_LENGTH added
//
// Revision 1.1  2000/08/20 15:23:18  Thomas Nussbaumer
// initial version
//
//
//
